<?php

namespace Drupal\seo_urls\Controller;

use Drupal\Component\Utility\Xss;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\seo_urls\SeoUrlManagerInterface;
use Drupal\taxonomy\TermInterface;
use Drupal\taxonomy\VocabularyInterface;

/**
 * Provides route responses for taxonomy.module.
 */
class SeoUrlCreateController extends ControllerBase {

  /**
   * SEO Url manager interface.
   *
   * @var \Drupal\seo_urls\SeoUrlManagerInterface
   */
  protected SeoUrlManagerInterface $seoUrlManager;

  /**
   * Retrieves the SEO Url manager interface.
   *
   * @return \Drupal\seo_urls\SeoUrlManagerInterface
   *   SEO Url manager interface.
   */
  protected function seoUrlManager(): SeoUrlManagerInterface {
    if (!isset($this->seoUrlManager)) {
      $this->seoUrlManager = \Drupal::service('seo_urls.manager');
    }
    return $this->seoUrlManager;
  }

  /**
   * Returns a form to add a new term to a vocabulary.
   *
   * @return array
   *   The taxonomy term add form.
   */
  public function addForm() {
    // Init SEU Url manager.
    $this->seoUrlManager();

    /** @var \Drupal\taxonomy\TermInterface $term */
    $term = $this->entityTypeManager()->getStorage($this->seoUrlManager::ENTITY_TYPE)->create(['vid' => $this->seoUrlManager::ENTITY_BUNDLE]);

    // Set the default canonical Url.
    if ($redirect_path = $this->getRedirectDestination()->get()) {
      $term->set($this->seoUrlManager::CANONICAL_URL_FIELD, [
        'uri' => 'internal:' . $this->seoUrlManager->clearPathPrefix($redirect_path),
        'title' => '',
        'options' => [],
      ]);
    }

    return $this->entityFormBuilder()->getForm($term);
  }

  /**
   * Route title callback.
   *
   * @param \Drupal\taxonomy\VocabularyInterface $taxonomy_vocabulary
   *   The vocabulary.
   *
   * @return string
   *   The vocabulary label as a render array.
   *
   * @deprecated in drupal:9.3.0 and is removed from drupal:10.0.0. There is no
   *   replacement; route title callbacks are internal.
   *
   * @see https://www.drupal.org/project/drupal/issues/3037157
   */
  public function vocabularyTitle(VocabularyInterface $taxonomy_vocabulary) {
    @trigger_error(__METHOD__ . ' is deprecated in drupal:9.3.0 and is removed from drupal:10.0.0. There is no replacement; route title callbacks are internal. See https://www.drupal.org/project/drupal/issues/3037157', E_USER_DEPRECATED);
    return [
      '#markup' => $taxonomy_vocabulary->label(),
      '#allowed_tags' => Xss::getHtmlTagList(),
    ];
  }

  /**
   * Route title callback.
   *
   * @param \Drupal\taxonomy\TermInterface $taxonomy_term
   *   The taxonomy term.
   *
   * @return array
   *   The term label as a render array.
   */
  public function termTitle(TermInterface $taxonomy_term) {
    return [
      '#markup' => $taxonomy_term->getName(),
      '#allowed_tags' => Xss::getHtmlTagList(),
    ];
  }

  /**
   * Access.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   Whether the current content type is allowed to create SEO Url.
   */
  public function access() {
    $router_match = \Drupal::routeMatch();
    $parameters = $router_match->getParameters()->all();
    $allowed_entity_types = $this->seoUrlManager()->getAllowedEntityTypes();
    $filtered_parameters = array_intersect_key($parameters, array_fill_keys($allowed_entity_types, TRUE));
    return AccessResult::allowedIf(!empty($filtered_parameters) || $router_match->getRouteName() === 'seo_urls.create');
  }

}
